import re
import json
import os
import urllib
import datetime
import shutil
import requests


import itchiodl.utils


class Game:
    """Representation of a game download"""

    def __init__(self, data):
        self.data = data["game"]
        self.name = self.data["title"]
        self.publisher = self.data["user"]["username"]
        self.link = self.data["url"]
        if "game_id" in data:
            self.id = data["id"]
            self.game_id = data["game_id"]
        else:
            self.id = False
            self.game_id = self.data["id"]

        matches = re.match(r"https://(.+)\.itch\.io/(.+)", self.link)
        self.game_slug = matches.group(2)
        self.publisher_slug = matches.group(1)

        self.files = []
        self.downloads = []

    def load_downloads(self, token):
        """Load all downloads for this game"""
        self.downloads = []
        if self.id:
            r = requests.get(
                f"https://api.itch.io/games/{self.game_id}/uploads?download_key_id={self.id}",
                headers={"Authorization": token},
            )
        else:
            r = requests.get(
                f"https://api.itch.io/games/{self.game_id}/uploads",
                headers={"Authorization": token},
            )
        j = r.json()
        for d in j["uploads"]:
            self.downloads.append(d)

    def download(self, token, platform):
        """Download a singular file"""
        print("Downloading", self.name)

        # if os.path.exists(f"{self.publisher_slug}/{self.game_slug}.json"):
        #    print(f"Skipping Game {self.name}")
        #    return

        self.load_downloads(token)

        if not os.path.exists(self.publisher_slug):
            os.mkdir(self.publisher_slug)

        if not os.path.exists(f"{self.publisher_slug}/{self.game_slug}"):
            os.mkdir(f"{self.publisher_slug}/{self.game_slug}")

        for d in self.downloads:
            if (
                platform is not None
                and d["traits"]
                and f"p_{platform}" not in d["traits"]
            ):
                print(f"Skipping {self.name} for platform {d['traits']}")
                continue
            self.do_download(d, token)

        with open(f"{self.publisher_slug}/{self.game_slug}.json", "w") as f:
            json.dump(
                {
                    "name": self.name,
                    "publisher": self.publisher,
                    "link": self.link,
                    "itch_id": self.id,
                    "game_id": self.game_id,
                    "itch_data": self.data,
                },
                f,
                indent=2,
            )

    def do_download(self, d, token):
        """Download a single file, checking for existing files"""
        print(f"Downloading {d['filename']}")

        file = itchiodl.utils.clean_path(d["filename"] or d["display_name"] or d["id"])
        path = itchiodl.utils.clean_path(f"{self.publisher_slug}/{self.game_slug}")

        if os.path.exists(f"{path}/{file}"):
            print(f"File Already Exists! {file}")
            if os.path.exists(f"{path}/{file}.md5"):

                with open(f"{path}/{file}.md5", "r") as f:
                    md5 = f.read().strip()

                    if md5 == d["md5_hash"]:
                        print(f"Skipping {self.name} - {file}")
                        return
                    print(f"MD5 Mismatch! {file}")
            else:
                md5 = itchiodl.utils.md5sum(f"{path}/{file}")
                if md5 == d["md5_hash"]:
                    print(f"Skipping {self.name} - {file}")

                    # Create checksum file
                    with open(f"{path}/{file}.md5", "w") as f:
                        f.write(d["md5_hash"])
                    return
                # Old Download or corrupted file?
                corrupted = False
                if corrupted:
                    os.remove(f"{path}/{file}")
                    return

            if not os.path.exists(f"{path}/old"):
                os.mkdir(f"{path}/old")

            print(f"Moving {file} to old/")
            timestamp = datetime.datetime.now().strftime("%Y-%m-%d")
            print(timestamp)
            shutil.move(f"{path}/{file}", f"{path}/old/{timestamp}-{file}")

        # Get UUID
        r = requests.post(
            f"https://api.itch.io/games/{self.game_id}/download-sessions",
            headers={"Authorization": token},
        )
        j = r.json()

        # Download
        if self.id:
            url = (
                f"https://api.itch.io/uploads/{d['id']}/"
                + f"download?api_key={token}&download_key_id={self.id}&uuid={j['uuid']}"
            )
        else:
            url = (
                f"https://api.itch.io/uploads/{d['id']}/"
                + f"download?api_key={token}&uuid={j['uuid']}"
            )
        # response_code = urllib.request.urlopen(url).getcode()
        try:
            itchiodl.utils.download(url, path, self.name, file)
        except itchiodl.utils.NoDownloadError:
            print("Http response is not a download, skipping")

            with open("errors.txt", "a") as f:
                f.write(
                    f""" Cannot download game/asset: {self.game_slug}
                    Publisher Name: {self.publisher_slug}
                    Path: {path}
                    File: {file}
                    Request URL: {url}
                    This request failed due to a missing response header
                    This game/asset has been skipped please download manually
                    ---------------------------------------------------------\n """
                )

            return
        except urllib.error.HTTPError as e:
            print("This one has broken due to an HTTP error!!")

            with open("errors.txt", "a") as f:
                f.write(
                    f""" Cannot download game/asset: {self.game_slug}
                    Publisher Name: {self.publisher_slug}
                    Path: {path}
                    File: {file}
                    Request URL: {url}
                    Request Response Code: {e.code}
                    Error Reason: {e.reason}
                    This game/asset has been skipped please download manually
                    ---------------------------------------------------------\n """
                )

            return

        # Verify
        if itchiodl.utils.md5sum(f"{path}/{file}") != d["md5_hash"]:
            print(f"Failed to verify {file}")
            return

        # Create checksum file
        with open(f"{path}/{file}.md5", "w") as f:
            f.write(d["md5_hash"])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['itchiodl', 'itchiodl.bundle_tool', 'itchiodl.downloader']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.10.0,<5.0.0',
 'clint>=0.5.1,<0.6.0',
 'requests>=2.28.1,<3.0.0']

entry_points = \
{'console_scripts': ['itch-download = itchiodl.downloader.__main__:main',
                     'itch-load-bundle = itchiodl.bundle_tool.__main__:main']}

setup_kwargs = {
    'name': 'itchiodl',
    'version': '2.1.2',
    'description': 'Python Scripts for downloading / archiving your itchio library',
    'long_description': '## IMPORTANT NOTICE: up until 2022-03-09, the package was called itchio, it is now called itchiodl (to avoid pypi conflicts)\n\n# Itchio Downloader Tool\n## Install\n```bash\npip install itchiodl\n```\n## Download All Games in library from account\n**Please Note:** Having too many jobs may cause rate-limiting from some testing 8 works fine but 10 starts giving errors.\n\n```bash\n# via python\npython -m itchiodl.downloader\n\n# via setup-tools entry point\nitch-download\n```\n\nThis uses the same API the itchio app uses to download the files. If you have 2FA enabled, generate an API key [here](https://itch.io/user/settings/api-keys) and run the following instead\n\n```bash\n# via python (with 4 concurrent downloads)\npython -m itchiodl.downloader --api-key=KEYHERE --jobs=4\n\n# via setup-tools entry point\nitch-download -k KEYHERE\n\n# download with multiple threads\nitch-download -k KEYHERE -j 4\n\n# only download osx or cross platform downloads\nitch-download -p osx\n```\n\n## Add All Games in a bundle to your library\n\n```bash\n# via python\npython -m itchiodl.bundle_tool\n\n# via setup-tools entry point\nitch-load-bundle\n```\n\nThis is a bit of a bodge, but it works. It essentially goes through and clicks the "Download" link on every item on the bundle\'s page, which adds it to your itchio library. It does not download any files. You will need the download page\'s URL (this will be in the bundle\'s email, and possibly your purchase history). It will not work with 2FA, and I\'m unlikely to be able to fix it without making it far more complicated\n\n\n## Errors\nif a download fails it will be reported in ```errors.txt``` in the same directory as your downloads\n\nAn example of which could look something like this:\n```Cannot download game/asset: <Game Name>\nPublisher Name: <Publisher Name>\nOutput File: <Publisher Name>/<Game Name>/<Specific Item>\nRequest URL: <Some URL>\nRequest Response Code: 404\nError Reason: Not Found\nThis game/asset has been skipped please download manually\n---------------------------------------------------------\n```\n\nThis is not a perfect solution but does prevent the whole process from crashing\n',
    'author': 'Peter Taylor',
    'author_email': 'me@et1.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
